// ntfsattribute.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include "ntfsattribute.h"
#include "ntfsutil.h"
#include "ntfsvolume.h"

NTFSAttribute::NTFSAttribute()
{
	vcnStart = vcnStop = sizeAlloc = sizeReal = sizeInitialized = 0;
	type = recordLength = dataLength = 0;
	flags = attributeId = dataOffset = compressionUnitSize = 0;
	nonResident = nameLength = 0;
	name = QString::null;
	data = NULL;
}

void NTFSAttribute::copy(NTFSAttribute * a)
{
	// Copy fields
	vcnStart = a->vcnStart;
	vcnStop = a->vcnStop;
	sizeAlloc = a->sizeAlloc;
	sizeReal = a->sizeReal;
	sizeInitialized = a->sizeInitialized;
	type = a->type;
	attributeId = a->attributeId;
	dataOffset = a->dataOffset;
	compressionUnitSize = a->compressionUnitSize;
	nonResident = a->nonResident;
	nameLength = a->nameLength;
	name = a->name;

	// Copy runlist
	runList.copy(&a->runList);

	// TODO: copy data
	if (!nonResident) WARN("Incomplete attribute copy!");
	data = NULL;
}

NTFSAttribute::~NTFSAttribute()
{
	if (data) delete data;
}

bool NTFSAttribute::read(NTFSStream * s)
{
	// Read type
	type = s->read32();
	//TRACE(QString("Reading attribute with type %1.").arg(toHex(type)));

	if (type == 0xFFFFFFFF)
	{
		// End-of-attributes descriptor
		// Even if we're the end of the attribute, we can count as valid, so return true.
		return true;
	}

	// Read attribute length
	recordLength = s->read32();
	// See if recordLength makes sense
	if (recordLength < 0x18)
	{
		// We're not valid, and we shouldn't keep reading.
		SWARN(QString("Invalid record length (%1): less than minimum.").arg(recordLength),s);
		return false;
	}

	// Read non-resident flag
	nonResident = s->read8();
	// See if non-resident flag makes sense
	if (nonResident != 0 && nonResident != 1)
	{
		// Nonsensical.
		SWARN(QString("Non-resident flag is invalid. (Got %1.)").arg(to8Hex(nonResident)),s);
		return false;
	}

	if (recordLength > s->getLength())
	{
		// We're not valid, and we shouldn't keep reading.
		SWARN(QString("Invalid record length for attribute (%1): larger than stream (%2).")
			.arg(toHex(recordLength)).arg(toHex(s->getLength())),s);
		return false;
	}

	// Read name length (TODO: validate?)
	nameLength = s->read8();
	// Read name offset
	nameOffset = s->read16();
	// Validate name offset
	if (nameOffset != 0x0000 && nameOffset != 0x0018 && nameOffset != 0x0040 && nameOffset != 0x0048)
	{
		// Name offset is invalid.
		SWARN(QString("Name offset (%1) is invalid.").arg(to16Hex(nameOffset)), s);
		return false;
	}

	// Read standard fields with no validation
	flags = s->read16();
	attributeId = s->read16();

	// Calculate the address of the end of the fixed-size header
	NT16 endOfFixedHeader = (flags & 0x0001 ? 0x0048 : 0x0040);

	// Next fields depend on resident/nonresident
	if (nonResident)
	{
		// Non-resident attribute
		// Get VCNs
		vcnStart = s->read64();
		vcnStop = s->read64();

		// Read offset to data runs, and verify that it makes sense
		dataOffset = s->read16();

		// dataOffset should be 0x40/0x48 + 2 * nameLength
		if (dataOffset > endOfFixedHeader + 2 * (nameLength + 1))
		{
			// Data offset is higher than expected/
			STRACE(QString("Unknown data offset detected. (Read %1, should be %2).")
					.arg(to16Hex(dataOffset))
					.arg(to16Hex((NT16)(endOfFixedHeader + 2 * nameLength))), s);
		}
		if (dataOffset < endOfFixedHeader + 2 * nameLength)
		{
			// Data offset doesn't make sense.
			SWARN(QString("Unknown data offset detected. (Read %1, should be %2).")
					.arg(to16Hex(dataOffset))
					.arg(to16Hex((NT16)(endOfFixedHeader + 2 * nameLength))), s);
			// We can't deal with this situation.
			//return false;
		}

		// Read compression unit size
		compressionUnitSize = s->read16();

		// Ignore padding
		s->read32();

		// Read size variables
		sizeAlloc = s->read64();
		sizeReal = s->read64();
		sizeInitialized = s->read64();
	}
	else
	{
		// Resident attribute
		// Read fields
		dataLength = s->read32();
		dataOffset = s->read16();

		// Discard indexed field (NT8) and padding (NT8)
		s->read16();
	}

	// Read the attribute's name, if any
	name = s->readString(nameLength);

	// Read attribute data
	if (nonResident)
	{
		// Non-resident
		// Calculate size of data runs
		dataLength = recordLength - dataOffset;

		// Read data runs into a local holder
		NTFSStream myS;
		s->getSubStream(&myS, dataOffset, dataLength);
		if (!runList.read(&myS))
		{
			SWARN("Failed to read data runs.", s);
		}
		myS.detach();
	}
	else
	{
		// Resident
		// Seek to offset
		s->seek(dataOffset);

		// Create a Data* object
		if (data) delete data;
		data = newDataByType(type);

		// Tell it to read the data chunk
		NTFSStream dataStream;
		s->getSubStream(&dataStream, dataOffset, dataLength);
		if (!data || !data->read(&dataStream))
		{
			SWARN("Failed to read data.", s);
		}
	}

	// Seek to end-of-record (which we should already be at)
	s->seek(recordLength);

	// We got here, so everything seems to have worked correctly
	return true;
}

QString NTFSAttribute::toString()
{
	QString aName, aContents, aFlags = "";

	if (name.length())
	{
		aFlags.append(QString("name \"%1\"").arg(name));
	}
	if (nonResident)
	{
		if (aFlags.length()) aFlags.append(", ");
		aFlags = QString("non-resident: %1").arg(runList.toString());
	}

	if (aFlags.length())
	{
		aFlags = QString(" [%1]").arg(aFlags);
	}

	if (data)
	{
		aName = data->dataType();
		aContents = data->toString();
	}
	else
	{
		aName = dataTypeToString(type);
	}

	// Fill aContents with "length = _" if necessary
	if (aContents.isNull() || aContents.isEmpty())
	{
		if (nonResident)
			aContents = QString("length = %1 / %2").arg(toHex(sizeReal)).arg(toHex(sizeAlloc));
		else
			aContents.sprintf("length = %i", dataLength);
	}

	// Fill aString with "aName (0x____): contents flags"
	QString aString;
	aString = QString("\t%1 (%2)%3%4%5")
		.arg(aName)
		.arg(toHex(type))
		.arg(aContents.length() > 0 ? ": " : "")
		.arg(aContents)
		.arg(aFlags);

	return aString;
}

// Returns true if the attribute is resident or has been made resident; false otherwise
bool NTFSAttribute::deReference(NTFSVolume * v)
{
	// Sanity checks
	if (!nonResident)
	{
		// We're already local
		// No de-referencing is needed
		return true;
	}
	if (v == NULL)
	{
		// Ack!
		BUG("Passed a NULL volume!");
		return false;
	}

	// See if the attribute is in only one chunk
	if (vcnStart != 0 || vcnStop != ((sizeAlloc / v->getVolInfo()->bytesPerCluster) - 1))
	{
		// In more than one chunk
		ERROR("De-referencing chunked non-resident attributes is currently unsupported.");
		TRACE(QString("vcnStart = %1, vcnStop = %2, sizeAlloc = %3")
			.arg(toHex(vcnStart)).arg(toHex(vcnStop)).arg(toHex(sizeAlloc)));
		return false;
	}

	// runList contains data runs and s ought to be an empty stream
	// Tell runList to copy its data into s, giving it a volume pointer, a
	// pointer to our target stream, the expected size of the attribute, and
	// its flags
	NTFSStream s;
	if (!runList.copyDataToStream(v, &s, sizeReal, flags))
	{
		// Failure :-(
		ERROR("Unable to retrieve contents of runlist.");
		return false;
	}

	// Create a Data* object
	if (data) delete data;
	data = newDataByType(type);

	// Tell it to read the data chunk
	if (!data->read(&s))
	{
		WARN("Failed to read attribute data.");
		delete data;
		return false;
	}

	// The data is in memory, so we're no longer resident
	nonResident = false;

	// Zero non-resident size values
	vcnStart = vcnStop = sizeAlloc = sizeReal = sizeInitialized = 0;

	// Update resident size values
	dataLength = s.getLength();
	dataOffset = 0;

	// Success!
	return true;
}

// De-references the attribute, writing directly to a file descriptor
bool NTFSAttribute::writeContentsToFile(NTFSVolume * v, int fd)
{
	// Are we resident?
	if (!nonResident)
	{
		// Uh oh... we might not have the data stream any more.
		if (type == 0x80 || type == 0x90 || type == 0xA0)
		{
			// We're an NTFSAttributeDataContainer, so we still have a stream
			NTFSAttributeDataContainer * d;
			d = (NTFSAttributeDataContainer *) data;
			return d->s.writeToFile(fd);
		}

		ERROR("Unable to write contents of a resident attribute.");
		return false;
	}
	else
	{
		// We're a non-resident attribute

		// Make sure we have a volume
		if (v == NULL)
		{
			// Ack!
			BUG("Passed a NULL volume!");
			return false;
		}

		// De-reference to file
		return runList.copyDataToFile(v, fd, sizeReal, flags);
	}
}

// Makes s point to our data
bool NTFSAttribute::getData(NTFSVolume * v, NTFSStream * s)
{
	// Are we resident?
	if (!nonResident)
	{
		// Uh oh... we might not have the data stream any more.
		if (type == 0x80 || type == 0x90 || type == 0xA0)
		{
			// We're an NTFSAttributeDataContainer, so we still have a stream
			NTFSAttributeDataContainer * d;
			d = (NTFSAttributeDataContainer *) data;

			if (d == NULL)
			{
				BUG("Attribute.data is NULL, expecting NTFSAttributeDataContainer!");
				return false;
			}

			d->s.getSubStream(s, 0, d->s.getLength());
			return true;
		}

		ERROR("Unable to write contents of a resident attribute.");
		return false;
	}
	else
	{
		// We're a non-resident attribute
		// Make sure we have a volume
		if (v == NULL)
		{
			// Ack!
			BUG("Passed a NULL volume!");
			return false;
		}

		// De-reference to stream
		return runList.copyDataToStream(v, s, sizeReal, flags);
	}
}
