// ntfsattributedata.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include "ntfsattributedata.h"
#include "ntfsutil.h"

QString dataTypeToString(NT32 type)
{
	switch (type)
	{
	case 0x10:
		return "$STANDARD_INFORMATION";
	case 0x20:
		return "$ATTRIBUTE_LIST";
	case 0x30:
		return "$FILE_NAME";
	case 0x40:
		return "$OBJECT_ID";
	case 0x50:
		return "$SECURITY_DESCRIPTOR";
	case 0x60:
		return "$VOLUME_NAME";
	case 0x70:
		return "$VOLUME_INFORMATION";
	case 0x80:
		return "$DATA";
	case 0x90:
		return "$INDEX_ROOT";
	case 0xA0:
		return "$INDEX_ALLOCATION";
	case 0xB0:
		return "$BITMAP";
	default:
		return "(Unknown)";
	}
}

NTFSAttributeData * newDataByType(NT32 type)
{
	switch (type)
	{
	case 0x10:
		return new NTFSAttributeDataStdInfo;
	case 0x20:
		return new NTFSAttributeDataAttribList;
	case 0x30:
		return new NTFSAttributeDataFilename;
	case 0x40:
		return new NTFSAttributeDataObjectId;
	case 0x50:
		return new NTFSAttributeDataSecurityDescriptor;
	case 0x60:
		return new NTFSAttributeDataVolumeName;
	case 0x70:
		return new NTFSAttributeDataVolumeInfo;
	case 0x80:
		return new NTFSAttributeDataData;
	case 0x90:
		return new NTFSAttributeDataIndexRoot;
	case 0xA0:
		return new NTFSAttributeDataIndexAlloc;
	case 0xB0:
		return new NTFSAttributeDataBitmap;
	default:
		return NULL;
	}
}

bool NTFSAttributeDataContainer::read(NTFSStream * in_s)
{
	in_s->getSubStream(&s, 0, in_s->getLength());
	s.seek(0);
	return true;
};


//////////////////////////////////////////////////////////////////////////////
// 0x10 $STANDARD_INFORMATION
//////////////////////////////////////////////////////////////////////////////
bool NTFSAttributeDataStdInfo::read(NTFSStream * s)
{
	if (s->getLength() - s->getOffset() < 48)
	{
		SERROR("Stream is too short to contain attribute!", s);
		return false;
	}

	cTime = s->read64();
	mTime = s->read64();
	aTime = s->read64();
	rTime = s->read64();
	fileFlags = s->read32();
	maxVersion = s->read32();
	version = s->read32();
	classId = s->read32();

	// Do we have extended fields?
	if (s->getLength() > s->getOffset() + 24)
	{
		// Yes
		// TODO: infer minimum volume version of 3.0 from this
		ownerId = s->read32();
		securityId = s->read32();
		quotaCharged = s->read64();
		updateSequenceNumber = s->read64();
	}
	else
	{
		// No extended fields
		ownerId = securityId = 0;
		quotaCharged = updateSequenceNumber = 0;
	}

	return true;
}

QString NTFSAttributeDataStdInfo::toString() const
{
	QString out;

	if (fileFlags & 0x0002) out.append("hidden ");
	if (fileFlags & 0x0004) out.append("system ");
	if (fileFlags & 0x0200) out.append("sparse ");
	if (fileFlags & 0x0800) out.append("compressed ");
	if (fileFlags & 0x4000) out.append("encrypted ");
	if (fileFlags & 0x0001) out.append("read-only ");

	if (out.length() == 0) out = "normal ";

	out.append("file");

	if (maxVersion > 0) out.append(QString(", version %1").arg(version));

	return out;
}

//////////////////////////////////////////////////////////////////////////////
// 0x20 $ATTRIBUTE_LIST
//////////////////////////////////////////////////////////////////////////////
bool NTFSAttributeDataAttribList::read(NTFSStream * in_s)
{
	// Copy stream
	return in_s->getSubStream(&s, in_s->getOffset(), in_s->getLength() - in_s->getOffset());
}

//////////////////////////////////////////////////////////////////////////////
// 0x30 $FILE_NAME
//////////////////////////////////////////////////////////////////////////////
bool NTFSAttributeDataFilename::read(NTFSStream * s)
{
	// Read fields
	parentRef = s->read64();
	cTime = s->read64();
	mTime = s->read64();
	aTime = s->read64();
	rTime = s->read64();
	allocSize = s->read64();
	realSize = s->read64();
	flags = s->read32();
	extendedField = s->read32();

	NT8 nameLen = s->read8();
	nameSpace = s->read8();
	fileName = s->readString(nameLen);

	return true;
}

QString NTFSAttributeDataFilename::toString() const
{
	return QString("\"%1\", parent ref %2")
		.arg(fileName).arg(toHexRef(parentRef));
}

//////////////////////////////////////////////////////////////////////////////
// 0x40 $OBJECT_ID
//////////////////////////////////////////////////////////////////////////////
bool NTFSAttributeDataObjectId::read(NTFSStream * in_s)
{
	guid = in_s->readGuidAsString();
	return true;
}

//////////////////////////////////////////////////////////////////////////////
// 0x50 $SECURITY_DESCRIPTOR
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// 0x60 $VOLUME_NAME
//////////////////////////////////////////////////////////////////////////////
bool NTFSAttributeDataVolumeName::read(NTFSStream * in_s)
{
	name = in_s->readString(in_s->getLength() / 2);
	return true;
}

//////////////////////////////////////////////////////////////////////////////
// 0x70 $VOLUME_INFORMATION
//////////////////////////////////////////////////////////////////////////////
bool NTFSAttributeDataVolumeInfo::read(NTFSStream * in_s)
{
	in_s->seek(8);
	major = in_s->read8();
	minor = in_s->read8();
	flags = in_s->read16();
	return true;
}

QString NTFSAttributeDataVolumeInfo::toString() const
{
	return QString("version %1.%2, %3%4")
		.arg(major).arg(minor)
		.arg((flags & 0x0001) ? "dirty" : "clean")
		.arg((flags & 0x8000) ? ", modified by CHKDSK" : "");
}

//////////////////////////////////////////////////////////////////////////////
// 0x80 $DATA
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// 0x90 $INDEX_ROOT
//////////////////////////////////////////////////////////////////////////////
QString NTFSAttributeDataIndexRoot::toString() const
{
	return QString("index length = %1 bytes").arg(toHex(s.getLength() - 0x20));
}

//////////////////////////////////////////////////////////////////////////////
// 0xA0 $INDEX_ALLOCATION
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// 0xB0 $BITMAP
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// 0xC0 $REPARSE_POINT
//////////////////////////////////////////////////////////////////////////////
bool NTFSAttributeDataReparsePoint::read(NTFSStream * in_s)
{
	flags = in_s->read32();
	return true;
}

QString NTFSAttributeDataReparsePoint::toString() const
{
	QString out;

	// TODO: fix flag logic
	switch (flags)
	{
	case 0x20000000:
		out = "Alias";
		break;
	case 0x68000005:
		out = "NSS";
		break;
	case 0x68000006:
		out = "NSS recover";
		break;
	case 0x68000007:
		out = "SIS";
		break;
	case 0x68000008:
		out = "DFS";
		break;
	case 0x88000003:
		out = "Mount point";
		break;
	case 0xA8000004:
		out = "HSM";
		break;
	case 0xE8000000:
		out = "Symlink";
		break;
	default:
		out = "Unknown";
		break;
	}
	if (flags & 0x40000000) out += ", high latency";
	if (flags & 0x80000000) out += ", Microsoft";

	return out;
}

//////////////////////////////////////////////////////////////////////////////
// 0xD0 $EA_INFORMATION
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// 0xE0 $EA
//////////////////////////////////////////////////////////////////////////////

