// ntfsattributedata.h
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#ifndef NTFSATTRIBUTEDATA_H
#define NTFSATTRIBUTEDATA_H

#include <qstring.h>

#include "ntfs.h"
#include "ntfsstream.h"

// Abstract base class
class NTFSAttributeData
{
public:
	virtual ~NTFSAttributeData() {};

	virtual QString dataType() const = 0;;
	virtual bool read(NTFSStream * in_s) { return true; };
	virtual QString toString() const { return QString::null; };
};

// Derived class
class NTFSAttributeDataContainer : public NTFSAttributeData
{
public:
	virtual ~NTFSAttributeDataContainer() {};
	virtual bool read(NTFSStream * in_s);
	NTFSStream s;
};

// Returns a new NTFSAttributeData * corresponding to type
NTFSAttributeData * newDataByType(NT32 type);
QString dataTypeToString(NT32 type);

//////////////////////////////////////////////////////////////////////////////
// 0x10 $STANDARD_INFORMATION
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataStdInfo : public NTFSAttributeData
{
public:
	virtual QString dataType() const { return "$STANDARD_INFORMATION"; };
	virtual bool read(NTFSStream * in_s);
	virtual QString toString() const;

	NT64 cTime;
	NT64 mTime;
	NT64 aTime;
	NT64 rTime;
	NT32 fileFlags;
	NT32 maxVersion;
	NT32 version;
	NT32 classId;
	NT32 ownerId;
	NT32 securityId;
	NT64 quotaCharged;
	NT64 updateSequenceNumber;
};

//////////////////////////////////////////////////////////////////////////////
// 0x20 $ATTRIBUTE_LIST
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataAttribList : public NTFSAttributeData
{
public:
	virtual QString dataType() const { return "$ATTRIBUTE_LIST"; };
	virtual bool read(NTFSStream * in_s);

	NTFSStream s;
};

//////////////////////////////////////////////////////////////////////////////
// 0x30 $FILE_NAME
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataFilename : public NTFSAttributeData
{
public:
	virtual QString dataType() const { return "$FILE_NAME"; };
	virtual bool read(NTFSStream * in_s);
	virtual QString toString() const;

	NT64 parentRef;
	NT64 cTime;
	NT64 mTime;
	NT64 aTime;
	NT64 rTime;
	NT64 allocSize;
	NT64 realSize;
	NT32 flags;
	NT32 extendedField;
	NT8 nameSpace;
	QString fileName;
};

//////////////////////////////////////////////////////////////////////////////
// 0x40 $OBJECT_ID
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataObjectId : public NTFSAttributeData
{
public:
	virtual QString dataType() const { return "$OBJECT_ID"; };
	virtual bool read(NTFSStream * in_s);
	virtual QString toString() const { return guid; };

	QString guid;
};

//////////////////////////////////////////////////////////////////////////////
// 0x50 $SECURITY_DESCRIPTOR
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataSecurityDescriptor : public NTFSAttributeDataContainer
{
public:
	virtual QString dataType() const { return "$SECURITY_DESCRIPTOR"; };
};

//////////////////////////////////////////////////////////////////////////////
// 0x60 $VOLUME_NAME
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataVolumeName : public NTFSAttributeData
{
public:
	virtual QString dataType() const { return "$VOLUME_NAME"; };
	virtual bool read(NTFSStream * in_s);
	virtual QString toString() const { return QString("\"%1\"").arg(name); };

	QString name;
};

//////////////////////////////////////////////////////////////////////////////
// 0x70 $VOLUME_INFORMATION
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataVolumeInfo : public NTFSAttributeData
{
public:
	virtual QString dataType() const { return "$VOLUME_INFORMATION"; };
	virtual bool read(NTFSStream * in_s);
	virtual QString toString() const;

	NT8 major;
	NT8 minor;
	NT16 flags;
};

//////////////////////////////////////////////////////////////////////////////
// 0x80 $DATA
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataData : public NTFSAttributeDataContainer
{
public:
	virtual QString dataType() const { return "$DATA"; };
};

//////////////////////////////////////////////////////////////////////////////
// 0x90 $INDEX_ROOT
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataIndexRoot : public NTFSAttributeDataContainer
{
public:
	virtual QString dataType() const { return "$INDEX_ROOT"; };
	virtual QString toString() const;
};

//////////////////////////////////////////////////////////////////////////////
// 0xA0 $INDEX_ALLOCATION
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataIndexAlloc : public NTFSAttributeDataContainer
{
public:
	virtual QString dataType() const { return "$INDEX_ALLOCATION"; };
};

//////////////////////////////////////////////////////////////////////////////
// 0xB0 $BITMAP
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataBitmap : public NTFSAttributeDataContainer
{
public:
	virtual QString dataType() const { return "$BITMAP"; };
};

//////////////////////////////////////////////////////////////////////////////
// 0xC0 $REPARSE_POINT
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataReparsePoint : public NTFSAttributeData
{
public:
	virtual QString dataType() const { return "$REPARSE_POINT"; };
	virtual bool read(NTFSStream * in_s);
	virtual QString toString() const;

	NT32 flags;
};

//////////////////////////////////////////////////////////////////////////////
// 0xD0 $EA_INFORMATION
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataEAInfo : public NTFSAttributeDataContainer
{
public:
	virtual QString dataType() const { return "$EA_INFORMATION"; };
};

//////////////////////////////////////////////////////////////////////////////
// 0xE0 $EA
//////////////////////////////////////////////////////////////////////////////
class NTFSAttributeDataEA : public NTFSAttributeDataContainer
{
public:
	virtual QString dataType() const { return "$EA"; };
};

#endif
