// ntfsrunlist.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include "ntfsrunlist.h"
#include "ntfsutil.h"
#include "ntfsvolumeinfo.h"
#include "ntfsvolume.h"

// For write()
#include <unistd.h>

NTFSRunList::NTFSRunList()
{
	// Tell list that it owns everything
	m_runList.setAutoDelete(true);
}

NTFSRunList::~NTFSRunList()
{
}

void NTFSRunList::copy(NTFSRunList * src)
{
	m_runList.clear();

	for (RUNLIST * cur = src->m_runList.first(); cur; cur = src->m_runList.next())
	{
		RUNLIST * n = new RUNLIST;
		n->vcn = cur->vcn;
		n->length = cur->length;
		n->lcn = cur->lcn;
		n->sparse = cur->sparse;

		m_runList.append(n);
	}
}

// Appends a runlist to the current runlist
void NTFSRunList::append(NTFSRunList * data)
{
	NT64 vcnOffset = 0;
	if (m_runList.count()) vcnOffset = m_runList.last()->vcn + m_runList.last()->length;

	for (RUNLIST * cur = data->m_runList.first(); cur; cur = data->m_runList.next())
	{
		RUNLIST * n = new RUNLIST;
		n->vcn = cur->vcn + vcnOffset;
		n->length = cur->length;
		n->lcn = cur->lcn;
		n->sparse = cur->sparse;

		m_runList.append(n);
	}
}

QString NTFSRunList::toString()
{
	if (m_runList.count() == 0) return "(empty runlist)";

	QString out;
	RUNLIST * cur;
	for (cur = m_runList.first(); cur; cur = m_runList.next())
	{
		if (out.length() > 0) out += ", ";
		out += QString("%1 cluster%2 at %3")
			.arg((NT32)cur->length).arg(cur->length == 1 ? "" : "s").arg(toHex(cur->lcn));
	}
	return out;
}

bool NTFSRunList::read(NTFSStream * s)
{
	// Local storate
	RUNLIST * cur = NULL;
	union {
		struct {
			unsigned length : 4;
			unsigned offset : 4;
		} pair;
		NT8 byte;
	} length;
	NT64 pos = 0;
	NT8 i;
	NT64 temp;
	NT64 offsetTemp;
	NT64 previousLcn = 0;

	// Read stream, terminating on 0x00
	while ((length.byte = s->read8()) != 0x00)
	{
		// Check that everything makes sense
		if (length.pair.offset > 8)
		{
			SERROR(QString("Unable to handle %1-byte offset.").arg(length.pair.offset), s);
			return false;
		}
		if (length.pair.length > 8)
		{
			SERROR(QString("Unable to handle %1-byte length.").arg(length.pair.length), s);
			return false;
		}

		// Create a new RUNLIST
		cur = new RUNLIST;

		// Set base VCN to the current position
		cur->vcn = pos;

		// Read length
		cur->length = 0;
		for (i = 0; i < length.pair.length; i ++)
		{
			// Read byte
			temp = s->read8();

			// Shift left i bytes
			temp = temp << (i * 8);

			// OR with current value
			cur->length |= temp;
		}

		// Should we read offset?
		if (length.pair.offset == 0)
		{
			// Offset is zero bytes; i.e. file is sparse
			cur->sparse = true;
			cur->lcn = 0;
		}
		else
		{
			// Read offset field
			offsetTemp = 0;
			for (i = 0; i < length.pair.offset; i ++)
			{
				temp = s->read8();
				temp = temp << (i * 8);
				offsetTemp |= temp;
			}

			// offsetTemp now contains the offset
			// However, since it's a variable-length signed value, it has to be
			// sign-extended from its length to 64 bits

			// Figure out the sign bit by shifting right (length * 8) - 1 bits
			NT8 signBit = offsetTemp >> ((length.pair.offset * 8) - 1);

			// Make sure we're rational
			if (signBit != (signBit & 0x01)) SBUG("What?!", s);

			// If we're negative, create a bitmask and OR with the current value
			if (signBit)
			{
				// Make a bitmask of all ones
				NT64 bitMask = 0xFFFFFFFFFFFFFFFF;

				// Zero the lowest length.pair.offset bytes
				bitMask = bitMask << (length.pair.offset * 8);

				// OR with the offset
				offsetTemp |= bitMask;
			}

			// offsetTemp now contains a signed value
			// Cast it to one, add it to the previous LCN, and stuff it in our
			// current RUNLIST.
			cur->lcn = previousLcn + ((NT64S)offsetTemp);
			
			// Store current LCN for use with later runlists
			previousLcn = cur->lcn;
		}

		// Sigh.

		// Add runlist to storage
		m_runList.append(cur);
	}

	// Everything seems fine
	return true;
}

bool NTFSRunList::first(NT64 * lcn, NT64 * length)
{
	RUNLIST * cur = m_runList.first();
	if (cur)
	{
		if (lcn) *lcn = cur->lcn;
		if (length) *length = cur->length;
		return true;
	}
	return false;
}

bool NTFSRunList::next(NT64 * lcn, NT64 * length)
{
	RUNLIST * cur = m_runList.next();
	if (cur)
	{
		if (lcn) *lcn = cur->lcn;
		if (length) *length = cur->length;
		return true;
	}
	return false;
}

// De-reference and put it in a stream
bool NTFSRunList::copyDataToStream(NTFSVolume * v, NTFSStream * s, NT64 len, NT16 flags)
{
	// Sanity checks
	if (v == NULL)
	{
		BUG("Called with NULL volume!");
		return false;
	}

	// Get information about what we're dealing with
	int fragments = m_runList.count();
	const NTFSVolumeInfo * vi = v->getVolInfo();

	// See if we're in zero chunks
	if (fragments == 0)
	{
		// No data -- make an empty stream
		char a = 0;
		s->attachCopy(&a, 0, 0);
		return true;
	}

	// See if we're in one chunk
	if (fragments == 1)
	{
		// Easy, efficient case
		RUNLIST * rl = m_runList.getFirst();
		NT64 rlLen = rl->length * vi->bytesPerCluster;

		// Stream the LCN to len
		if (!v->streamData(s, vi->getPositionForLcn(rl->lcn), MIN(rlLen, len)))
		{
			ERROR(QString("Unable to read from LCN %1!").arg(toHex(rl->lcn)));
			return false;
		}

		s->seek(0);

		// Everything is peachy
		return true;
	}

	// More than one chunk

	// Sanity check
	// Don't load a 1.5+ GB attribute into memory
	if (len > 1536 * 1024 * 1024)
	{
		ERROR("Refusing to load a fragmented runlist into memory over 1.5 GB in size.");
		return false;
	}

	// Read into memory

	// Memory chunk
	char * out = new char[len];

	// Position within chunk
	NT64 curPos = 0;

	// Amount we're going to read total
	NT64 lenRemaining = len;

	NTFSStream curStream;
	NT64 lenRunlist;
	NT32 curLen;
	for (RUNLIST * cur = m_runList.first(); cur; cur = m_runList.next())
	{
		// The length of the current runlist entry
		lenRunlist = cur->length * vi->bytesPerCluster;

		// See if it's too big for our tastes
		if (lenRunlist > 1024 * 1024 * 1024)
		{
			ERROR("Refusing to load single runlist entry over 1 GB in size.");
			return false;
		}

		// The length we're going to read from the current runlist
		curLen = MIN(lenRunlist, lenRemaining);

		// Stream the runlist up to to curLen
		if (!v->streamData(&curStream, vi->getPositionForLcn(cur->lcn), curLen))
		{
			ERROR(QString("Unable to read from LCN %1!").arg(toHex(cur->lcn)));
			delete[] out;
			return false;
		}

		// curStream contains our data -- have it memcpy into our holder
		curStream.readArray(out + curPos, curLen);

		// Increase current position by the length we just added to it
		curPos += curLen;

		// Decrease our length remaining by subtracting
		lenRemaining -= curLen;

		// Bail out if we're done reading
		if (curPos >= len)
			break;
	}

	// Attach the stream to our data
	s->attach(out, curPos, 0);

	return (curPos >= len);
}

// De-reference and write it directly to a file descriptor
bool NTFSRunList::copyDataToFile(NTFSVolume * v, int fd, NT64 len, NT16 flags)
{
	// Sanity checks
	if (v == NULL)
	{
		BUG("Called with NULL volume!");
		return false;
	}

	// Get information about what we're dealing with
	int fragments = m_runList.count();
	const NTFSVolumeInfo * vi = v->getVolInfo();

	// Holder stream
	NTFSStream s;

	// See if we're in zero chunks
	if (fragments == 0)
	{
		// No data
		// Bail out, returning true if we expected no data, false otherwise
		return (len == 0);
	}

	// Position within chunk to read
	NT64 curPos = 0;

	// Amount we're going to read total
	NT64 lenRemaining = len;

	NT64 lenRunlist;
	NT64 curLen;
	for (RUNLIST * cur = m_runList.first(); cur && lenRemaining; cur = m_runList.next())
	{
		// The length of the current runlist entry
		lenRunlist = cur->length * vi->bytesPerCluster;

		// The length we're writing (MIN(lenRemaining, lenRunlist))
		curLen = (lenRemaining > lenRunlist ? lenRunlist : lenRemaining);

		// Write LCN/length to fd
		if (!v->copyData(vi->getPositionForLcn(cur->lcn), curLen, fd))
		{
			ERROR(QString("Unable to read from LCN %1!").arg(toHex(cur->lcn)));
			return false;
		}

		// Increase current position by the length we just added to it
		curPos += curLen;

		// Decrease our length remaining by subtracting
		lenRemaining -= curLen;

		// Bail out if we're done reading
		if (curPos >= len)
			break;
	}

	// We're done reading requested length or out of runlist entries
	// Return true if we read everything they wanted
	return (curPos >= len);
}

void NTFSRunList::addSparseRun(NT64 vcn, NT64 length)
{
	RUNLIST * n = new RUNLIST;
	n->vcn = vcn;
	n->length = length;
	n->lcn = 0;
	n->sparse = true;
	m_runList.append(n);
}
