// ntfsstream.h
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#ifndef NTFSSTREAM_H
#define NTFSSTREAM_H

#include "ntfs.h"
#include "ntfsvolumeinfo.h"

#include <qstring.h>
#include <qcstring.h>
#include <qptrlist.h>

class NTFSVolume;

class NTFSStream;
class NTFSStream
{
public:
	friend class NTFSVolume;

    NTFSStream();
    ~NTFSStream();

	// Data manipulation
	//////////////////////////////////////////////////////////////////////////
	// Attach the stream to a chunk of data
	// If parent is set, the Volume is assumed to own the data (from mmap()ing
	// it). If parent is not set, data must have been allocated with new[],
	// and subsequently becomes property of the stream.
	void attach(char * data, NT32 length, NT64 position, NTFSVolume * parent = NULL);
	// Attach the stream to a chunk of data
	// The data will be copied into NTFSStream's own, personal chunk of data.
	void attachCopy(const char * data, NT32 length, NT64 position);
	// Detach the stream's chunk of data
	// This involves freeing memory.
	void detach();
	// Makes an NTFSStream have its own copy of the data
	void unbind();

	// Substream creation
	//////////////////////////////////////////////////////////////////////////
	// Assigns a stream to a substream of this one directly (changes in one affect the other)
	bool getSubStream(NTFSStream *target, NT32 offset, NT32 length) const;
	// Assigns a stream to a substream of this one indirectly (changes are independent)
	bool getSubStreamCopy(NTFSStream *target, NT32 offset, NT32 length) const;

	// Position functions
	//////////////////////////////////////////////////////////////////////////
	// Seeks a stream to "offset", relative to the beginning of the stream
	inline void seek(NT32 offset) { m_cursor = (offset > m_dataLen ? m_dataLen : offset); };
	// Returns the cursor location relative to the beginning of the stream
	inline NT32 getOffset() const { return m_cursor; };
	// Returns the cursor location as a position on the underlying NTFSVolume
	inline NT64 getVolumePosition() const { return (NT64)getOffset() + m_position; };
	// Returns the length of the stream's data
	inline NT32 getLength() const { return m_dataLen; };

	// Returns the cursor location as a record number on the underlying NTFSVolume
	inline NT64 getVolumeRecord() const
		{ return (m_volInfo ? m_volInfo->getRecordForPosition(getVolumePosition()) : 0); };
	// Returns the cursor location as an offset from a record number on the underlying NTFSVolume
	inline NT64 getVolumeRecordOffset() const
		{ return (m_volInfo ? m_volInfo->getOffsetFromRecord(getVolumePosition()) : 0); };

	// Returns the cursor location as an LCN on the underlying NTFSVolume
	inline NT64 getVolumeLcn() const
		{ return (m_volInfo ? m_volInfo->getLcnForPosition(getVolumePosition()) : 0); };
	// Returns the cursor location as an offset from an LCN on the underlying NTFSVolume
	inline NT64 getVolumeLcnOffset() const
		{ return (m_volInfo ? m_volInfo->getOffsetFromLcn(getVolumePosition()) : 0); };

	// Update operation
	//////////////////////////////////////////////////////////////////////////
	// "Patches" an offset with new data (for use with update sequences, etc.)
	// Operates on the current stream and any explicitly shared streams
	void patch(NT32 offset, const char * data, NT32 length);

	// Reading of unsigned data types
	//////////////////////////////////////////////////////////////////////////
	// Reads an NT8 from the stream
	NT8 read8();
	// Reads an NT16 from the stream
	NT16 read16();
	// Reads an NT32 from the stream
	NT32 read32();
	// Reads an NT64 from the stream
	NT64 read64();

	// Reading of signed data types
	//////////////////////////////////////////////////////////////////////////
	// Reads an NT8S from the stream
	inline NT8S read8s() { return (NT8S)read8(); };
	// Reads an NT16S from the stream
	inline NT16S read16s() { return (NT16S)read16(); };
	// Reads an NT32S from the stream
	inline NT32S read32s() { return (NT32S)read32(); };
	// Reads an NT64S from the stream
	inline NT64S read64s() { return (NT64S)read64(); };

	// Reading of special data types
	//////////////////////////////////////////////////////////////////////////
	// Reads a Unicode string with known length from the stream
	QString readString(NT16 length);
	// Reads an ASCII string with known length from the stream
	QString readAscii(NT16 length);
	// Reads a GUID and formats MS registry style
	QString readGuidAsString();
	// Reads directly into a caller-allocated array (fast!)
	bool readArray(char * dest, NT32 length);
	// Reads into a QByteArray, resizing as necessary
	bool readByteArray(QByteArray &a, NT32 length);

	// Copying of stream data
	//////////////////////////////////////////////////////////////////////////
	// Write data into file descriptor "fd", ignoring the cursor
	bool writeToFile(int fd, NT32 length) const;
	inline bool writeToFile(int fd) const { return writeToFile(fd, this->getLength()); };

	// Determining of volume information
	//////////////////////////////////////////////////////////////////////////
	// Get sector size, or guess if not available
	inline NT16 getVolumeBytesPerSector() const { return (m_volInfo ? m_volInfo->bytesPerSector : 512); };
	// Get volume information
	inline const NTFSVolumeInfo * getVolumeInfo() const { return m_volInfo; };

protected:
	NTFSVolumeInfo * m_volInfo;

	char * m_data;
	NT32 m_dataLen;
	NT32 m_cursor;

	NT64 m_position;

	// Parent volume management
	NTFSVolume * m_volume;
};

#endif
