// ntfs.cpp
// Part of SalvageNTFS
// Copyright (C) Will Glynn
//////////////////////////////////////////////////////////////////////////////
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html

#include <qstring.h>
#include "ntfs.h"
#include "ntfsutil.h"
#include "ntfsstream.h"

// Fixed-length
QString to8Hex(NT8 n)
{
	QString o;
	o.sprintf("0x%02X", n);
	return o;
}

// Fixed-length
QString to16Hex(NT16 n)
{
	QString o;
	o.sprintf("0x%04X", n);
	return o;
}

// Fixed-length
QString to32Hex(NT32 n)
{
	QString o;
	o.sprintf("0x%08X", n);
	return o;
}

// Fixed-length
QString to64Hex(NT64 n)
{
        // Split an NT64 into two NT32s
        union {
                NT64 n64;
                NT32 n32[2];
        } conv;
        conv.n64 = n;

		QString out;
		out.sprintf("0x%08X%08X", conv.n32[1], conv.n32[0]);
		return out;
}

// Variable-length
QString toHex(NT32 n)
{
	if (n == 0) return "0x00";

	QString o;
	o.sprintf("%X", n);
	if (o.length() % 2 != 0)
		return QString("0x0%1").arg(o);
	else
		return QString("0x%1").arg(o);
}

// Variable-length
QString toHex(NT64 n)
{
	// Split an NT64 into two NT32s
	union {
		NT64 n64;
		NT32 n32[2];
	} conv;
	conv.n64 = n;

	QString o1, o2;
	o1.sprintf("%X", conv.n32[1]);
	o2.sprintf("%X", conv.n32[0]);

	if (conv.n32[1] == 0)
	{
		o1 = o2;
	}
	else
	{
		// o1 is not empty; prepend zeros to o2 and append o2 to o1
		while (o2.length() < 8) o2.prepend('0');
		o1.append(o2);
	}
	if (o1.length() == 0) return "0x00";

	if (o1.length() % 2 != 0)
		return QString("0x0%1").arg(o1);
	else
		return QString("0x%1").arg(o1);
}

// Format refNum as a reference number (0x19AF#5)
QString toHexRef(NT64 refNum)
{
	return QString("%1#%2").arg(toHex(refNum & 0xFFFFFFFFFFFF)).arg((NT16)(refNum >> 48));
}

NT8 hexCharTo8(QChar c)
{
	switch (c.latin1())
	{
		case '1':
			return 1;
		case '2':
			return 2;
		case '3':
			return 3;
		case '4':
			return 4;
		case '5':
			return 5;
		case '6':
			return 6;
		case '7':
			return 7;
		case '8':
			return 8;
		case '9':
			return 9;
		case 'A':
			return 10;
		case 'B':
			return 11;
		case 'C':
			return 12;
		case 'D':
			return 13;
		case 'E':
			return 14;
		case 'F':
			return 15;
		default:
		case '0':
			return 0;
	}
}

NT64 toNum(QString string)
{
	// Easy case
	if (string.length() == 0) return 0;

	// Are we hexidecimal?
	if (string.left(2) == "0x")
	{
		// Hex

		// Get upper-case data section of string
		QString data = string.upper().mid(2, string.length() - 2);

		// Make un-hex
		NT64 num = 0;
		for (unsigned int i = 0; i < data.length(); i ++)
		{
			// Shift number one position left
			num = num << 4;

			// Get low nibble and OR it with num
			num |= hexCharTo8(data[i]);
		}

		return num;
	}
	else
	{
		// Decimal
		return atoll(string.latin1());
	}
}

time_t time64ToUnix(NT64 timestamp)
{
	// Function inspired by:
	//   http://support.microsoft.com/default.aspx?scid=KB;en-us;q167296
	return (time_t)((timestamp - 116444736000000000) / 10000000);
}

void bugTrace(QString file, int line, QString function, QString prefix, QString message, const NTFSStream * s)
{
	nTrace(file, line, function, prefix, message, s);
	abort();
}

void nTrace(QString file, int line, QString function, QString prefix, QString message, const NTFSStream * s)
{
	QString suffix = "";
	if (s) suffix = QString(" (record %1 @ %2)")
		.arg(toHex(s->getVolumeRecord())).arg(toHex(s->getVolumeRecordOffset()));

	fprintf(stderr, "%s[%s:%i] %s(): %s%s\n",
		prefix.latin1(),
		file.latin1(),
		line,
		function.latin1(),
		message.latin1(),
		suffix.latin1());
	fflush(stderr);
}
